/* Tactical Analysis for AIs
   Copyright (C) 2005 Eric A. McDonald.

Xconq is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.  See the file COPYING.  */

/*! \file 
    \brief Tactical Analysis for AIs

    Part of the AI API, Level 3. 

    Provides tactical level AI implementation.

    \note Nothing in this file should be required for AI implementation.

*/

#include "conq.h"
#include "kernel.h"
#include "aiscore.h"
#include "aiunit.h"
#include "aiunit2.h"
#include "aitact.h"

namespace Xconq {
namespace AI {

/** Task Inspection **/

int
n_plan_to_construct_at(Side *side, int x, int y, short *p_utypes)
{
    int n = 0;
    Unit *unit = NULL;
    Task *agenda = NULL;
    Side *side2 = NULL;
    int u2 = NONUTYPE;

    assert_error(side,
		 "AI Task Inspector: Attempted to access a NULL side");
    assert_warning_return(inside_area(x, y),
			  "AI Task Inspector: Inquiry about out-of-arena cell",
			   0);
    for_all_sides(side2) {
	if ((side2 != side) && !allied_side(side2, side))
	    continue;
	for_all_side_units(side2, unit) {
	    agenda = (unit->plan ? unit->plan->tasks : NULL);
	    if (!agenda)
		continue;
	    // If movement involved, see what next task is.
	    if ((agenda->type == TASK_MOVE_TO) || (agenda->type == TASK_OCCUPY))
		agenda = agenda->next;
	    if (!agenda)
		continue;
	    if (agenda->type != TASK_CONSTRUCT)
		continue;
	    // Skip, if construction to occur in transport.
	    if (agenda->args[2] != -1)
		continue;
	    // If construction location matches location of interest...
	    if ((agenda->args[3] == x) && (agenda->args[4] == y)) {
		if (p_utypes) {
		    for_all_unit_types(u2) {
			if (p_utypes[u2] && (u2 == agenda->args[0])) {
			    ++n;
			    break;
			}
		    }
		}
		else
		    ++n;
	    }
	}
    }
    return n;
}

/** Resource Contention Assessment **/

int
resource_contention_with_any(int u, int x, int y, Side *side)
{
    static short *p_advanced;

    Side *side2 = NULL;
    Unit *unit2 = NULL;
    int ux = -1, uy = -1, dist = -1;
    int userid = -1;
    int u2 = NONUTYPE;

    assert_error(is_unit_type(u),
"AI Resource Contention Assessor: Encountered invalid unit type");
    assert_error(side,
"AI Resource Contention Assessor: Attempted to access a NULL side");
    assert_error(inside_area(x, y),
"AI Resource Contention Assessor: Attempted to access resource contention " \
"at an out-of-area location");

    // Initialize advanced utypes array, if necessary.
    // TODO: Should already have this precalculated somewhere.
    if (!p_advanced) {
	p_advanced = (short *)xmalloc(numutypes * sizeof(short));
	for_all_unit_types(u2)
	    p_advanced[u2] = u_advanced(u2);
    }
    // Inspect all friendly units for contention.
    for_all_sides(side2) {
	if ((side2 != side) && !allied_side(side2, side))
	    continue;
	// Is anyone using or will be using the cell?
	if (user_defined() && u_advanced(u)) {
	    // TODO: Should consider reach of highest immobile utype on 
	    //	utype's evolutionary ladder.
	    for_all_cells_within_range(x, y, u_reach(u), ux, uy) {
		if (!inside_area(ux, uy))
		    continue;
		if (n_plan_to_construct_at(side2, ux, uy, p_advanced))
		    return TRUE;
		userid = user_at(ux, uy);
		if (NOUSER == userid)
		    continue;
		// TODO: Need user views so that we do not inspect enemy usage.
		unit2 = find_unit(userid);
		if (!in_play(unit2))
		    continue;
		if (!allied_side(unit2->side, side))
		    continue;
		return TRUE;
	    } // for all cells within range
	}
	// Are min separation distance reqs met?
	for_all_side_units(side2, unit2) {
	    u2 = unit2->type;
	    dist = distance(x, y, unit2->x, unit2->y);
	    // TODO: Make separation distance be TableUU rather than gvar.
	    if (u_advanced(u) && u_advanced(u2)) {
		if (dist < g_ai_advanced_unit_separation())
		    return TRUE;
	    }
	}
    } // for all sides
    return FALSE;
}

/** Threat Assessment and Handling **/

/*!
    \todo Implement.
*/

TacticalConcern
handle_immediate_danger(Unit *unit)
{
    TacticalConcern taccon = TC_NONE;
    
    assert_error(in_play(unit), "Attempted to access an out-of-play unit");
    // TODO: Implement.
    return taccon;
}

/** Overall Tactical Assessment and Handling **/

/*!
    \todo Implement handling of occ danger.
    \todo Implement handling of targets of opportunity.
*/

TacticalConcern
handle_tactical_situation(Unit *unit)
{
    TacticalConcern taccon = TC_NONE;

    assert_error(in_play(unit), "Attempted to access an out-of-play unit");
    /* Threat Assessment */
    // Handle immediate danger to unit.
    if ((taccon = handle_immediate_danger(unit)))
	return taccon;
    // TODO: Handle immediate danger to unit's occs.
    /* Targets of Opportunity */
    // TODO: Implement.
    return taccon;
}

} // namespace Xconq::AI
} // namespace Xconq
